// Headers used in azure services.
pub const X_MS_DATE: &str = "x-ms-date";
pub const X_MS_VERSION: &str = "x-ms-version";
pub const CONTENT_MD5: &str = "content-md5";

// Env values used in azure services.
pub const AZURE_VERSION: &str = "2019-12-12";
